#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtGui, QtWidgets
import inspect
from builtins import int, float, str, bool
from typing import *


widget_mapping = {
    int: QtWidgets.QSpinBox,
    float: QtWidgets.QDoubleSpinBox,
    bool: QtWidgets.QCheckBox,
    str: QtWidgets.QLineEdit,
}

val_setters = {
    QtWidgets.QSpinBox: "setValue",
    QtWidgets.QDoubleSpinBox: "setValue",
    QtWidgets.QCheckBox: "setChecked",
    QtWidgets.QLineEdit: "setText",
}


class Arg(QtCore.QObject):
    acceptable_types = (int, float, str, bool)
    sig_changed = QtCore.pyqtSignal(object)

    def __init__(
            self,
            name: str,
            typ: type,
            val: Union[int, float, str, bool],
            parent: QtWidgets.QWidget,
            vlayout: QtWidgets.QVBoxLayout,
            tooltip: Optional[str] = None
    ):
        """
        Creates the appropriate QWidget interface.

        Parameters
        ----------
        name : str
            argument name

        typ : type
            function type, one of ``int``, ``float``, ``str`` or ``bool``.
            Used for determining the correct QWidget to be used

        val : Union[int, float, str, bool]
            default value for the widget

        parent : QtWidgets.QWidget
            parent widget

        vlayout : QtWidgets.QVBoxLayout
            parent VBoxLayout

        tooltip : str
            toolTip

        Attributes
        ----------
        sig_changed : object
            emits ``self.val`` when GUI value is changed.
        """
        super(Arg, self).__init__(parent)

        self.parent = parent
        self.vlayout = vlayout

        self.hlayout = QtWidgets.QHBoxLayout()

        self._qlabel = QtWidgets.QLabel(self.parent)
        self.hlayout.addWidget(self._qlabel)
        self.name = name

        self.typ = typ

        self.widget: QtWidgets.QWidget  #: QWidget for the argument
        self.widget = widget_mapping[self.typ](self.parent)
        self.hlayout.addWidget(self.widget)

        if typ is not str:
            self.hlayout.addSpacerItem(
                QtWidgets.QSpacerItem(
                    40, 20, QtWidgets.QSizePolicy.Expanding,
                    QtWidgets.QSizePolicy.Minimum
                )
            )

        self.val = val

        self.vlayout.addLayout(self.hlayout)

        if self.typ is str:
            self.widget.textEdited.connect(lambda v: setattr(self, '_val', v))
            self.widget.textEdited.connect(lambda: self.sig_changed.emit(self.val))
        elif self.typ is bool:
            self.widget.toggled.connect(lambda v: setattr(self, '_val', v))
            self.widget.toggled.connect(lambda: self.sig_changed.emit(self.val))

        if tooltip is not None:
            self._qlabel.setToolTip(tooltip)
            self.widget.setToolTip(tooltip)

    @property
    def name(self) -> str:
        """argument name"""
        return self._name

    @name.setter
    def name(self, n):
        self._name = n
        self._qlabel.setText(f'{self._name}: ')

    @property
    def val(self) -> Union[int, float, str, bool]:
        """current argument value"""
        return self._val

    @val.setter
    def val(self, v: Union[int, float, str, bool]):
        if v is None:
            self._val = v
            return

        assert isinstance(v, self.acceptable_types)

        self._val = v

        # use the correct func to set the value based on type
        getattr(self.widget, val_setters[type(self.widget)])(self.val)

    def __repr__(self):
        return f"name:\t{self.name}\n" f"val:\t{self.val}\n" f"typ:\t{self.typ}"


class ArgNumeric(Arg):
    acceptable_types = (int, float)

    def __init__(
            self,
            name: str,
            typ: type,
            val: Union[int, float],
            parent: QtWidgets.QWidget,
            vlayout: QtWidgets.QVBoxLayout,
            minmax: tuple = (-1, 999),
            step: Union[int, float] = 1,
            use_slider: bool = False,
            suffix: str = None,
            **kwargs
    ):
        """
        Creates numerical QWidget interface

        Parameters
        ----------
        minmax : tuple
            min & max values

        step : Union[int, float]
            step size for the spin box

        use_slider : Optional[bool]
            adds a slider below the spin box

        suffix : Optional[str]
            text suffix for the spin box, like data units

        **kwargs
            passed to Arg
        """
        super().__init__(name, typ, val, parent, vlayout, **kwargs)

        self.slider = None

        self.minmax = minmax

        self.step = step

        if use_slider:
            self.slider = QtWidgets.QSlider(self.parent)
            self.slider.setOrientation(QtCore.Qt.Horizontal)
            self.slider.setMaximum(self.max)
            self.slider.setMinimum(self.min)
            self.widget.valueChanged.connect(self.slider.setValue)
            self.slider.valueChanged.connect(self.widget.setValue)
            self.vlayout.addWidget(self.slider)

        self.suffix = suffix
        if self.suffix is not None:
            self.widget.setSuffix(self.suffix)

        self.widget.valueChanged.connect(lambda v: setattr(self, '_val', v))
        self.widget.valueChanged.connect(lambda: self.sig_changed.emit(self.val))
        self.val = val

    def set_slider(self):
        if self.slider is not None:
            self.slider.setMaximum(self.max)
            self.slider.setMinimum(self.min)

    @property
    def minmax(self) -> tuple:
        """minmax limits for the widget"""
        return tuple(self._minmax)

    @minmax.setter
    def minmax(self, minmax: tuple):
        self._minmax = list(minmax)
        self.min = self.minmax[0]
        self.max = self.minmax[1]

    @property
    def min(self) -> Union[int, float]:
        """min value limit for the widget"""
        return self._min

    @min.setter
    def min(self, v: Union[int, float]):
        assert isinstance(v, (int, float))
        self._min = v
        self._minmax[0] = v

        self.widget.setMinimum(v)
        self.set_slider()

    @property
    def max(self) -> Union[int, float]:
        """max value limit for the widget"""
        return self._max

    @max.setter
    def max(self, v: Union[int, float]):
        assert isinstance(v, (int, float))
        self._max = v
        self._minmax[1] = v

        self.widget.setMaximum(v)
        self.set_slider()

    @property
    def step(self) -> Union[int, float]:
        """step size for the widget"""
        return self._step

    @step.setter
    def step(self, v: Union[int, float]):
        assert isinstance(v, (int, float))
        self._step = v
        self.widget.setSingleStep(v)

    def __repr__(self):
        return (
            f"{super(ArgNumeric, self).__repr__()}\n"
            f"minmax:\t{self.minmax}\n"
            f"step:\t{self.step}\n"
            f"suffix:\t{self.suffix}"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtCore, QtGui, QtWidgets
import inspect
from typing import *
from collections import namedtuple
from functools import partial
from .argument import Arg, ArgNumeric


def _get_argument(sig: inspect.Parameter, parent, vlayout, **opts):
    if sig.default is inspect._empty:
        default = None
    else:
        default = sig.default

    kwargs = dict(
        name=sig.name,
        typ=sig.annotation,
        val=default,
        parent=parent,
        vlayout=vlayout
    )

    kwargs.update(opts)

    if sig.annotation in [int, float]:
        return ArgNumeric(**kwargs)

    else:
        return Arg(**kwargs)


class Function(QtCore.QObject):
    # emit the entire dict
    sig_changed = QtCore.pyqtSignal(dict)
    sig_set_clicked = QtCore.pyqtSignal(dict)

    # emits the name of the arg and its value
    sig_arg_changed = QtCore.pyqtSignal(str, object)

    def __init__(
            self,
            func: callable,
            arg_opts: dict = None,
            parent: Optional[QtWidgets.QWidget] = None,
            kwarg_entry: bool = False,
    ):
        """
        Creates a widget based on the function signature

        Parameters
        ----------
        func : callable
            A function with type annotations

        arg_opts : dict
            manually set certain features of an Arg

        parent : Optional[QtWidgets.QWidget]
            parent QWidget

        kwarg_entry : bool
            Not yet implemented.
            include a text box for kwargs entry


        Attributes
        -------
        sig_changed : dict
            Emitted when an argument value changes.
            Emits dict for all function arguments.
            See ``get_data()`` for details on the dict.

        sig_set_clicked : dict
            Emitted when the "Set" button is clicked.
            Emits dict for all function arguments.
            See ``get_data()`` for details on the dict.

        sig_arg_changed : str, object
            Emitted when specific argument value changes.
            Emits argument name and argument value


        Examples
        --------

        **Basic**

        .. code-block:: python
            :linenos:

            from PyQt5 import QtWidgets
            from qtap import Function

            # annotated function
            def f(a: int = 1, b: float = 3.14, c: str = 'yay', d: bool = True):
                pass

            app = QtWidgets.QApplication([])

            # basic
            func = Function(f)
            func.widget.show()

            app.exec()

        **Opt Args**

        .. code-block:: python
            :linenos:

            from PyQt5 import QtWidgets
            from qtap import Function
            from pyqtgraph.console import ConsoleWidget


            def f(a: int = 1, b: float = 3.14, c: str = 'yay', d: bool = True):
                pass


            if __name__ == '__main__':
                app = QtWidgets.QApplication([])

                # opt args dict
                opts = \
                    {
                        'b':
                            {
                                'use_slider': True,
                                'minmax': (0, 100),
                                'step': 1,
                                'suffix': '%',
                                'typ': int,
                                'tooltip': 'yay tooltips'
                            }
                    }

                func = Function(f, arg_opts=opts)
                func.widget.show()

                console = ConsoleWidget(parent=func.widget, namespace={'this': func})
                func.vlayout.addWidget(console)

                app.exec()


        """
        super(Function, self).__init__(parent)

        self.widget = QtWidgets.QWidget(parent)

        self.vlayout = QtWidgets.QVBoxLayout(self.widget)

        self.callable = func

        self.name = self.callable.__name__
        self._qlabel = QtWidgets.QLabel(self.widget)
        self._qlabel.setStyleSheet("font-weight: bold")
        self._qlabel.setText(self.name)
        self.vlayout.addWidget(self._qlabel)

        arg_names = inspect.signature(func).parameters.keys()
        arg_sigs = inspect.signature(func).parameters.values()

        self.arg_opts = dict.fromkeys(arg_names)
        self.arg_opts = {arg: {} for arg in arg_names}
        if arg_opts is not None:
            self.arg_opts.update(arg_opts)

        # Add all the arguments as named tuples
        # so they're accessible like attributes
        # dynamically named based on the args from the function!
        Arguments = namedtuple("Arguments", arg_names)
        self.arguments = Arguments(
            *(
                _get_argument(
                    sig,
                    parent=self.widget,
                    vlayout=self.vlayout,
                    **self.arg_opts[sig.name]
                )
                for sig in arg_sigs
            )
        )

        # button at the bottom, sends a "set" signal when clicked
        self.button_set = QtWidgets.QPushButton(self.widget)
        self.button_set.setText('Set')
        self.vlayout.addWidget(self.button_set)
        self.button_set.clicked.connect(
            partial(self._emit_data, self.sig_set_clicked)
        )

        for arg in self.arguments:
            # emit entire dict when arg is changed
            arg.sig_changed.connect(
                partial(self._emit_data, self.sig_changed)
            )

            # also emit just arg.name and arg.val when changed
            arg.sig_changed.connect(
                partial(self.sig_arg_changed.emit, arg.name)
            )

    def _emit_data(self, sig: QtCore.pyqtBoundSignal):
        sig.emit(self.get_data())

    def get_data(self) -> Dict[str, object]:
        """
        Get the data from the function arguments

        Returns
        -------
        dict
            dict keys are the argument names, dict values are the argument vals

        """

        return {arg.name: arg.val for arg in self.arguments}

    def set_data(self, d: dict):
        for arg in d.keys():
            getattr(self.arguments, arg).val = d[arg]

    def __repr__(self):
        return f'{self.name}' \
                   f'\n' + \
            '\n'.join(
                [
                    f'  {arg.name}:\n' \
                        f'    {arg.typ}\n' \
                        f'    {arg.val}'
                    for arg in self.arguments
                ]
            )


class Functions(QtWidgets.QWidget):
    sig_changed = QtCore.pyqtSignal(dict)
    sig_set_clicked = QtCore.pyqtSignal(dict)

    def __init__(
            self,
            functions: List[callable],
            arg_opts: Optional[List[dict]] = None,
            parent: Optional[QtWidgets.QWidget] = None,
            scroll: bool = False,
            orient: str = 'V',
            columns: bool = False,
            **kwargs
    ):
        """

        Parameters
        ----------
        functions : List[callable]
            list of functions

        arg_opts : List[dict], optional
            optional list of dicts to manually set features of an argument.
            passed to ``Function``

        parent : QtWidgets.QWidget, optional
            parent widget

        scroll : bool
            Not yet implemented

        orient : str
            orientation of the individual functions. One of ``V`` or ``H``.
            Default orientation is ``V`` (vertical)

        columns : bool
            Not yet implemented

        **kwargs
            passed to QtWidgets.QWidget.__init__()

        Attributes
        -------
        sig_changed : dict
            Emitted when an underlying function emits sig_changed().
            The emitted dict comes from ``get_data()``,
            see the docstring for ``get_data()`` for details.

        sig_set_clicked : dict
            Emitted when an underlying function emits sig_set_clicked().
            The emitted dict comes from ``get_data()``,
            see the docstring for ``get_data()`` for details.

        Examples
        --------

        **Basic**

        .. code-block:: python
            :linenos:

            from PyQt5 import QtWidgets
            from qtap import Functions
            from pyqtgraph.console import ConsoleWidget


            def func_A(a: int = 1, b: float = 3.14, c: str = 'yay', d: bool = True):
                pass


            def func_B(x: float = 50, y: int = 2.7, u: str = 'bah'):
                pass


            if __name__ == '__main__':
                app = QtWidgets.QApplication([])

                functions = Functions([func_A, func_B])

                console = ConsoleWidget(parent=functions, namespace={'this': functions})
                functions.main_layout.addWidget(console)

                functions.show()

                app.exec()

        **Opt Args**

        .. code-block:: python
            :linenos:

            from PyQt5 import QtWidgets
            from qtap import Functions
            from pyqtgraph.console import ConsoleWidget


            def func_A(a: int = 1, b: float = 3.14, c: str = 'yay', d: bool = True):
                pass


            def func_B(x: float = 50, y: int = 2.7, u: str = 'bah'):
                pass


            if __name__ == '__main__':
                app = QtWidgets.QApplication([])

                # opt args for ``func_A``
                opts_A = \
                    {
                        'b':
                            {
                                'use_slider': True,
                                'minmax': (0, 100),
                                'step': 1,
                                'suffix': '%',
                                'typ': int,
                                'tooltip': 'yay tooltips'
                            }
                    }

                # functions where one has ``opt_args``
                functions = Functions(
                    functions=[func_A, func_B],
                    arg_opts=[opts_A, None], # opt_args in same order as functions
                )

                console = ConsoleWidget(parent=functions, namespace={'this': functions})
                functions.main_layout.addWidget(console)

                functions.show()

                app.exec()

        """
        super().__init__(parent, **kwargs)

        _functions = namedtuple(
            'Functions',
            [f.__name__ for f in functions]
        )

        if arg_opts is None:
            arg_opts = [None] * len(functions)

        self.functions = _functions(
            *(
                Function(
                    func,
                    opt,
                    parent=self
                )
                for func, opt in zip(functions, arg_opts)
            )
        )

        if scroll:
            self.vlayout = QtWidgets.QVBoxLayout(self)

            self.scroll_area = QtWidgets.QScrollArea(self)
            self.vlayout.addWidget(self.scroll_area)
            self.scroll_area.setWidgetResizable(True)

            self.scroll_content = QtWidgets.QWidget(self.scroll_area)
            self.scroll_layout = QtWidgets.QVBoxLayout(self.scroll_content)
            self.scroll_content.setLayout(self.scroll_layout)

            self.main_layout = self.scroll_layout
        else:
            if orient in ['V', 'vertical']:
                self.main_layout = QtWidgets.QVBoxLayout(self)
            elif orient in ['H', 'horizontal']:
                self.main_layout = QtWidgets.QHBoxLayout(self)

        f: Function
        for f in self.functions:
            self.main_layout.addWidget(f.widget)

            # emit dict when any function changes
            f.sig_changed.connect(
                partial(self._emit_data, self.sig_changed)
            )

            # emit dict when any function is set
            f.sig_set_clicked.connect(
                partial(self._emit_data, self.sig_set_clicked)
            )

    def _emit_data(self, sig):
        sig.emit(self.get_data())

    def get_data(self) -> Dict[callable, dict]:
        """

        Returns
        -------
            dict
                dict keys are the functions, each dict values is a kwargs dict

        """
        return {f.callable: f.get_data() for f in self.functions}

    def __repr__(self):
        return '\n'.join(
            [
                f.__repr__() for f in self.functions
            ]
        )



from setuptools import setup

with open("readme-pypi.md", 'r') as fh:
    long_description = fh.read()

setup(
    name='qtap',
    version='0.1.1',
    packages=['qtap'],
    url='https://github.com/kushalkolar/qtap',
    license='GPL v3.0',
    author='Kushal Kolar',
    author_email='kushalkolar@alumni.ubc.ca',
    description='Automatic Qt parameter entry widgets using function signatures ',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.6',
)
